from unittest.mock import Mock, patch

import pytest

from django_qstash.tasks import shared_task


@shared_task
def sample_task(x, y):
    return x + y


@shared_task(name="custom_task", deduplicated=True)
def sample_task_with_options(x, y):
    return x * y


@pytest.mark.django_db
class TestQStashTasks:
    def test_basic_task_execution(self):
        """Test that tasks can be executed directly"""
        result = sample_task(2, 3)
        assert result == 5

    def test_task_with_options(self):
        """Test that tasks with custom options work"""
        result = sample_task_with_options(4, 5)
        assert result == 20

    @patch("django_qstash.tasks.qstash_client")
    def test_task_delay(self, mock_client):
        """Test that delay() sends task to QStash"""
        mock_response = Mock()
        mock_response.message_id = "test-id-123"
        mock_client.message.publish_json.return_value = mock_response

        result = sample_task.delay(2, 3)

        assert result.task_id == "test-id-123"
        mock_client.message.publish_json.assert_called_once()

    @patch("django_qstash.tasks.qstash_client")
    def test_task_apply_async(self, mock_client):
        """Test that apply_async() works with countdown"""
        mock_response = Mock()
        mock_response.message_id = "test-id-456"
        mock_client.message.publish_json.return_value = mock_response

        result = sample_task.apply_async(args=(2, 3), countdown=60)

        assert result.task_id == "test-id-456"
        call_kwargs = mock_client.message.publish_json.call_args[1]
        assert call_kwargs["delay"] == "60s"



import functools
from typing import Any, Callable, Dict, Optional, Tuple

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from qstash import QStash

QSTASH_TOKEN = getattr(settings, "QSTASH_TOKEN", None)
DJANGO_QSTASH_DOMAIN = getattr(settings, "DJANGO_QSTASH_DOMAIN", None)
DJANGO_QSTASH_WEBHOOK_PATH = getattr(
    settings, "DJANGO_QSTASH_WEBHOOK_PATH", "/qstash/webhook/"
)
if not QSTASH_TOKEN or not DJANGO_QSTASH_DOMAIN:
    raise ImproperlyConfigured("QSTASH_TOKEN and DJANGO_QSTASH_DOMAIN must be set")

# Initialize QStash client once
qstash_client = QStash(QSTASH_TOKEN)


class QStashTask:
    def __init__(
        self,
        func: Optional[Callable] = None,
        name: Optional[str] = None,
        delay_seconds: Optional[int] = None,
        deduplicated: bool = False,
        **options: Dict[str, Any],
    ):
        self.func = func
        self.name = name or (func.__name__ if func else None)
        self.delay_seconds = delay_seconds
        self.deduplicated = deduplicated
        self.options = options
        self.callback_domain = DJANGO_QSTASH_DOMAIN.rstrip("/")
        self.webhook_path = DJANGO_QSTASH_WEBHOOK_PATH.strip("/")

        if func is not None:
            functools.update_wrapper(self, func)

    def __get__(self, obj, objtype):
        """Support for instance methods"""
        return functools.partial(self.__call__, obj)

    def __call__(self, *args, **kwargs):
        """
        Execute the task, either directly or via QStash based on context
        """
        # Handle the case when the decorator is used without parameters
        if self.func is None:
            return self.__class__(
                args[0],
                name=self.name,
                delay_seconds=self.delay_seconds,
                deduplicated=self.deduplicated,
                **self.options,
            )

        # If called directly (not through delay/apply_async), execute the function
        if not getattr(self, "_is_delayed", False):
            return self.func(*args, **kwargs)

        # Reset the delayed flag
        self._is_delayed = False

        # Prepare the payload
        payload = {
            "function": self.func.__name__,
            "module": self.func.__module__,
            "args": args,  # Send args as-is
            "kwargs": kwargs,
            "task_name": self.name,
            "options": self.options,
        }

        # Ensure callback URL is properly formatted
        callback_domain = self.callback_domain
        if not callback_domain.startswith(("http://", "https://")):
            callback_domain = f"https://{callback_domain}"

        url = f"{callback_domain}/{self.webhook_path}/"
        # Send to QStash using the official SDK
        response = qstash_client.message.publish_json(
            url=url,
            body=payload,
            delay=f"{self.delay_seconds}s" if self.delay_seconds else None,
            retries=self.options.get("max_retries", 3),
            content_based_deduplication=self.deduplicated,
        )
        # Return an AsyncResult-like object for Celery compatibility
        return AsyncResult(response.message_id)

    def delay(self, *args, **kwargs) -> "AsyncResult":
        """Celery-compatible delay() method"""
        self._is_delayed = True
        return self(*args, **kwargs)

    def apply_async(
        self,
        args: Optional[Tuple] = None,
        kwargs: Optional[Dict] = None,
        countdown: Optional[int] = None,
        **options: Dict[str, Any],
    ) -> "AsyncResult":
        """Celery-compatible apply_async() method"""
        self._is_delayed = True
        if countdown is not None:
            self.delay_seconds = countdown
        self.options.update(options)

        # Fix: Ensure we're passing the arguments correctly
        args = args or ()
        kwargs = kwargs or {}
        return self(*args, **kwargs)


class AsyncResult:
    """Minimal Celery AsyncResult-compatible class"""

    def __init__(self, task_id: str):
        self.task_id = task_id

    def get(self, timeout: Optional[int] = None) -> Any:
        """Simulate Celery's get() method"""
        raise NotImplementedError("QStash doesn't support result retrieval")

    @property
    def id(self) -> str:
        return self.task_id


def shared_task(
    func: Optional[Callable] = None,
    name: Optional[str] = None,
    deduplicated: bool = False,
    **options: Dict[str, Any],
) -> QStashTask:
    """
    Decorator that mimics Celery's shared_task

    Can be used as:
        @shared_task
        def my_task():
            pass

        @shared_task(name="custom_name", deduplicated=True)
        def my_task():
            pass
    """
    if func is not None:
        return QStashTask(func, name=name, deduplicated=deduplicated, **options)
    return lambda f: QStashTask(f, name=name, deduplicated=deduplicated, **options)



from __future__ import annotations

from datetime import timedelta

from django.apps import apps
from django.conf import settings
from django.core.management.base import BaseCommand
from django.utils import timezone

DJANGO_QSTASH_RESULT_TTL = getattr(settings, "DJANGO_QSTASH_RESULT_TTL", 604800)


class Command(BaseCommand):
    help = f"""Clears stale task results older than\n
    {DJANGO_QSTASH_RESULT_TTL} seconds (settings.DJANGO_QSTASH_RESULT_TTL)"""

    def add_arguments(self, parser):
        parser.add_argument(
            "--no-input",
            action="store_true",
            help="Do not ask for confirmation",
        )
        parser.add_argument(
            "--since",
            type=int,
            help="The number of seconds ago to clear results for",
        )

    def handle(self, *args, **options):
        no_input = options["no_input"]
        since = options.get("since") or DJANGO_QSTASH_RESULT_TTL
        cutoff_date = timezone.now() - timedelta(seconds=since)
        try:
            TaskResult = apps.get_model("django_qstash_results", "TaskResult")
        except LookupError:
            self.stdout.write(
                self.style.ERROR(
                    "Django QStash Results not installed.\nAdd `django_qstash.results` to INSTALLED_APPS and run migrations."
                )
            )
            return
        to_delete = TaskResult.objects.filter(date_done__lt=cutoff_date)

        if not to_delete.exists():
            self.stdout.write("No stale Django QStash task results found")
            return

        # use input to confirm  deletion
        self.stdout.write(
            f"Deleting {to_delete.count()} task results older than {cutoff_date} ({DJANGO_QSTASH_RESULT_TTL} seconds)"
        )
        if not no_input:
            if input("Are you sure? (y/n): ") != "y":
                self.stdout.write("Skipping deletion")
                return

        deleted_count, _ = to_delete.delete()

        self.stdout.write(
            self.style.SUCCESS(f"Successfully deleted {deleted_count} stale results.")
        )



from __future__ import annotations

import logging

from django.apps import apps
from django.utils import timezone

logger = logging.getLogger(__name__)


def store_task_result(
    task_id, task_name, status, result=None, traceback=None, args=None, kwargs=None
):
    """Helper function to store task results if the results app is installed"""
    try:
        TaskResult = apps.get_model("django_qstash_results", "TaskResult")
        task_result = TaskResult.objects.create(
            task_id=task_id,
            task_name=task_name,
            status=status,
            date_done=timezone.now(),
            result=result,
            traceback=traceback,
            args=args,
            kwargs=kwargs,
        )
        return task_result
    except LookupError:
        # Model isn't installed, skip storage
        logger.debug(
            "Django QStash Results not installed. Add `django_qstash.results` to INSTALLED_APPS and run migrations."
        )
        return None
